/*
 * Copyright 2024 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package xyz.zhouxy.plusone.oss;

import java.util.Objects;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnClass;
import org.springframework.boot.context.properties.EnableConfigurationProperties;
import org.springframework.context.annotation.Configuration;
import org.springframework.util.StringUtils;

import io.minio.MinioAsyncClient;
import io.minio.MinioClient;
import xyz.zhouxy.plusone.commons.util.AssertTools;

/**
 * S3 配置
 *
 * @author <a href="http://zhouxy.xyz:3000/ZhouXY108}">ZhouXY</a>
 */
@Configuration
@EnableConfigurationProperties(S3Properties.class)
@ConditionalOnClass(value = {
        MinioClient.class,
        MinioAsyncClient.class,
})
public class PlusoneS3AutoConfiguration {

    private final OSSType ossType;
    private final Protocol protocol;
    private final Integer port;
    private final String endpoint;
    private final String accessKey;
    private final String secretKey;
    private final String region;

    private final boolean secure;

    @Autowired
    public PlusoneS3AutoConfiguration(S3Properties s3Properties) {
        this(s3Properties.getType(),
                s3Properties.getProtocol(),
                s3Properties.getEndpoint(),
                s3Properties.getPort(),
                s3Properties.isSecure(),
                s3Properties.getAccessKey(),
                s3Properties.getSecretKey(),
                s3Properties.getRegion());
    }

    private PlusoneS3AutoConfiguration(OSSType ossType,
            Protocol protocol, String endpoint, Integer port, boolean secure,
            String accessKey, String secretKey, String region) {
        AssertTools.checkArgument(Objects.nonNull(ossType), "Property \"ossType\" cannot be null");
        AssertTools.checkArgument(StringUtils.hasText(endpoint), "Property \"endpoint\" must has text.");
        AssertTools.checkArgument(StringUtils.hasText(accessKey), "Property \"accessKey\" must has text.");
        AssertTools.checkArgument(StringUtils.hasText(secretKey), "Property \"secretKey\" must has text.");
        AssertTools.checkArgument(StringUtils.hasText(region), "Property \"region\" must has text.");

        this.ossType = ossType;

        int i = endpoint.indexOf("://");
        if (i == -1) {
            // endpoint 不带协议
            this.protocol = protocol != null ? protocol : Protocol.HTTP;
            this.endpoint = endpoint;
        } else {
            // endpoint 带协议
            if (protocol != null) {
                // protocol 不为空，则使用 protocol
                this.protocol = protocol;
            } else {
                // protocol 为空，则使用 endpoint 中的协议
                String protocolInEndPoint = endpoint.substring(0, i);
                this.protocol = Protocol.of(protocolInEndPoint);
            }
            // 去除协议部分
            this.endpoint = endpoint.substring(i + "://".length());
        }
        this.port = port;
        this.secure = secure;
        this.accessKey = accessKey;
        this.secretKey = secretKey;
        this.region = region;
    }

    public MinioClient ossClient() {
        final MinioClient.Builder minioClientBuilder = MinioClient.builder();
        final String url = this.protocol.getCode() + "://" + this.endpoint;
        if (this.port != null) {
            minioClientBuilder.endpoint(url, this.port, this.secure);
        }
        else {
            minioClientBuilder.endpoint(url);
        }
        return minioClientBuilder
                .region(this.region)
                .credentials(this.accessKey, this.secretKey)
                .build();
    }

    /**
     * @return the ossType
     */
    public OSSType getOssType() {
        return ossType;
    }

    public Protocol getProtocol() {
        return protocol;
    }

    /**
     * @return the port
     */
    public Integer getPort() {
        return port;
    }

    public String getEndpoint() {
        return endpoint;
    }

    public String getAccessKey() {
        return accessKey;
    }

    public String getSecretKey() {
        return secretKey;
    }

    /**
     * @return the region
     */
    public String getRegion() {
        return region;
    }
}
